"""Dataclasses module"""

from dataclasses import dataclass, field
from pathlib import Path

from throttlebuster.constants import DownloadMode
from throttlebuster.helpers import get_duration_string, get_filesize_string


@dataclass(frozen=False)
class DownloadTracker:
    """Download part metadata"""

    url: str
    saved_to: Path
    index: int
    bytes_offset: int
    expected_size: int
    streaming_chunk_size: int = 0
    downloaded_size: int = 0
    download_mode: DownloadMode = DownloadMode.AUTO

    @property
    def is_complete(self) -> bool:
        """Checks whether the download was complete"""
        return self.downloaded_size >= self.expected_size

    @property
    def expected_size_string(self) -> str:
        return get_filesize_string(self.expected_size)

    @property
    def downloaded_size_string(self) -> str:
        return get_filesize_string(self.downloaded_size)

    def update_downloaded_size(self, new_chunk_size: int) -> int:
        """Updates the downloaded size value

        Args:
            new_chunk_size (int): Streaming chunk size

        Returns:
            int: New downloaded-size value
        """
        self.streaming_chunk_size = new_chunk_size
        self.downloaded_size += new_chunk_size
        return self.downloaded_size


@dataclass(frozen=True)
class DownloadedFile:
    """Complete downloaded file metadata"""

    url: str
    saved_to: Path
    expected_size: int
    size: int
    duration: float
    """Download time in seconds"""
    merge_duration: float
    file_parts: list[DownloadTracker] = field(default_factory=list)

    @property
    def tasks_used(self) -> int:
        return len(self.file_parts)

    @property
    def is_complete(self) -> bool:
        return self.expected_size == self.size

    @property
    def size_string(self) -> str:
        return get_filesize_string(self.size)

    @property
    def duration_string(self) -> str:
        return get_duration_string(self.duration)

    @property
    def merge_duration_string(self) -> str:
        return get_duration_string(self.merge_duration)

    @property
    def total_duration(self) -> float:
        return self.duration + self.merge_duration

    @property
    def total_duration_string(self) -> str:
        return get_duration_string(self.total_duration)



"""Commandline module"""

# TODO: Add click to cli extras
import logging
import os
import sys

import click

from throttlebuster.constants import (
    CURRENT_WORKING_DIR,
    DEFAULT_TASKS,
    DEFAULT_TASKS_LIMIT,
    DEFAULT_READ_TIMEOUT_ATTEMPTS,
    DOWNLOAD_PART_EXTENSION,
    DownloadMode,
)

DEBUG = os.getenv("BEGUG", "0") == "1"

command_context_settings = dict(auto_envvar_prefix="THROTTLEBUSTER")


def prepare_start(quiet: bool, verbose: bool) -> None:
    """Set up some stuff for better CLI usage such as:

    - Set higher logging level for some packages.
    ...

    """
    if verbose > 3:
        verbose = 2
    logging.basicConfig(
        format=("[%(asctime)s] : %(levelname)s - %(message)s" if verbose else "[%(module)s] %(message)s"),
        datefmt="%d-%b-%Y %H:%M:%S",
        level=(
            logging.ERROR
            if quiet
            # just a hack to ensure
            #           -v -> INFO
            #           -vv -> DEBUG
            else (30 - (verbose * 10))
            if verbose > 0
            else logging.INFO
        ),
    )
    # logging.info(f"Using host url - {HOST_URL}")
    packages = ("httpx",)
    for package_name in packages:
        package_logger = logging.getLogger(package_name)
        package_logger.setLevel(logging.WARNING)


@click.group()
@click.version_option(package_name="throttlebuster")
def throttlebuster():
    """Accelerate file downloads by overcoming common throttling restrictions
    envvar-prefix : THROTTLEBUSTER."""


@click.command(context_settings=command_context_settings)
@click.argument("url")
@click.option(
    "-T",
    "--tasks",
    type=click.IntRange(1, DEFAULT_TASKS_LIMIT),
    help="Number of tasks to carry out the download",
    default=DEFAULT_TASKS,
    show_default=True,
)
@click.option(
    "-Z",
    "--chunk-size",
    type=click.INT,
    help="Streaming download chunk size in kilobytes",
    default=256,
    show_default=True,
)
@click.option(
    "-D",
    "--dir",
    help="Directory for saving the downloaded file to",
    type=click.Path(exists=True, file_okay=False, writable=True, resolve_path=True),
    default=CURRENT_WORKING_DIR,
    show_default=True,
)
@click.option(
    "-P",
    "--part-dir",
    help="Directory for temporarily saving the downloaded file-parts to",
    type=click.Path(exists=True, file_okay=False, writable=True, resolve_path=True),
    default=CURRENT_WORKING_DIR,
    show_default=True,
)
@click.option(
    "-E",
    "--part-extension",
    help="Filename extension for download parts",
    default=DOWNLOAD_PART_EXTENSION,
    show_default=True,
)
@click.option(
    "-H",
    "--request-headers",
    help="Httpx request header - [key value] : default",
    nargs=2,
    multiple=True,
)
@click.option(
    "-C",
    "--request-cookies",
    help="Httpx request cookie - [key value]: default",
    nargs=2,
    multiple=True,
)
@click.option(
    "-B",
    "--merge-buffer-size",
    type=click.IntRange(1, 102400),
    default=256,
    help="Buffer size for merging the separated files in kilobytes",
    show_default=True,
)
@click.option("-F", "--filename", help="Filename for the downloaded content")
@click.option(
    "-M",
    "--mode",
    help="Whether to start or resume incomplete download",
    type=click.Choice(DownloadMode.map().keys(), case_sensitive=False),
    default=DownloadMode.AUTO.value,
    show_default=True,
)
@click.option("-L", "--file-size", type=click.INT, help="Size of the file to be downloaded")
@click.option(
    "-R",
    "--timeout-retry-attempts",
    type=click.INT,
    help="Number of times to retry download upon read request timing out",
    show_default=True,
    default=DEFAULT_READ_TIMEOUT_ATTEMPTS,
)
@click.option(
    "-K",
    "--colour",
    default="cyan",
    help="Progress bar display color",
    show_default=True,
)
@click.option(
    "-k",
    "--keep-parts",
    is_flag=True,
    help="Whether to retain the separate download parts",
)
@click.option(
    "-s",
    "--simple",
    is_flag=True,
    help="Show percentage and bar only in progressbar",
)
@click.option(
    "-t",
    "--test",
    is_flag=True,
    help="Just test if download is possible but do not actually download",
)
@click.option(
    "-a",
    "--ascii",
    is_flag=True,
    help="Use unicode (smooth blocks) to fill the progress-bar meter",
)
@click.option(
    "-l",
    "--no-leave",
    help="Do not keep all leaves of the progressbar",
    is_flag=True,
)
@click.option(
    "-z",
    "--disable-progress-bar",
    is_flag=True,
    help="Do not show progress_bar",
)
@click.option(
    "-i",
    "--suppress-incompatible-error",
    help="Do no raise error when response headers lack Etag",
    is_flag=True,
)
@click.option(
    "-q",
    "--quiet",
    is_flag=True,
    help="Do not show any interactive information",
)
@click.option(
    "-v",
    "--verbose",
    help="Show more detailed information",
    count=True,
    default=0,
    show_default=True,
)
def download_command(
    tasks: int,
    chunk_size: int,
    dir: str,
    part_dir: str,
    part_extension: str,
    request_headers: list[tuple[str]],
    request_cookies: list[tuple[str]],
    merge_buffer_size: int,
    quiet: bool,
    verbose: int,
    **run_kwargs,
):
    """Download file using http protocol"""
    prepare_start(quiet, verbose)

    from throttlebuster import ThrottleBuster

    throttlebuster = ThrottleBuster(
        dir=dir,
        chunk_size=chunk_size,
        tasks=tasks,
        part_dir=part_dir,
        part_extension=part_extension,
        merge_buffer_size=merge_buffer_size,
        request_headers=request_headers,
        cookies=list(request_cookies),
    )
    if quiet:
        run_kwargs["disable_progress_bar"] = True

    run_kwargs["leave"] = run_kwargs.get("no_leave") is False
    run_kwargs.pop("no_leave")
    run_kwargs["mode"] = DownloadMode.map().get(run_kwargs.get("mode"))

    throttlebuster.run_sync(**run_kwargs)


@click.command(context_settings=command_context_settings)
@click.argument("throttle", type=click.INT)
@click.option(
    "-U",
    "--url",
    help="Url to the target file",
)
@click.option("-S", "--size", type=click.INT, help="Size in bytes of the targeted file")
@click.option(
    "-T",
    "--tasks",
    help="Tasks amount to base the estimate on : Range (2-30)",
    type=click.IntRange(1, DEFAULT_TASKS_LIMIT),
)
@click.option("-j", "--json", is_flag=True, help="Stdout estimates in json format")
def estimate_command(throttle: int, url: str | None, size: int, tasks: int, json: bool):
    """Estimate download duration for different tasks"""
    assert size or url, "Either size of the file (--size) or url to it (--url) is required."

    import rich

    from throttlebuster.helpers import get_duration_string, get_filesize_string

    if size:
        size_in_bytes = size

    elif url:
        from throttlebuster import ThrottleBuster

        throttle_buster = ThrottleBuster()
        response = throttle_buster.run_sync(url, test=True)
        size_in_bytes = int(response.headers.get("content-length"))

    estimates: list[tuple[str]] = []

    def update_estimates(task: int):
        load_per_task = size_in_bytes / task
        download_duration = load_per_task / throttle
        download_duration_string = get_duration_string(download_duration)
        load_per_task_string = get_filesize_string(load_per_task)
        estimates.append((str(task), download_duration_string, load_per_task_string))

    if tasks is None:
        for task in range(1, 21):
            update_estimates(task)
        estimates.reverse()

    else:
        update_estimates(tasks)

    if json:
        rich.print_json(data=dict(estimates=estimates), indent=4)

    else:
        from rich.table import Table

        table = Table(
            "Tasks",
            "Duration",
            "Load per task",
            title=(f"{get_filesize_string(size_in_bytes)} at {get_filesize_string(throttle)}/s"),
            show_lines=False,
        )
        for row in estimates:
            table.add_row(*row)

        rich.print(table)


def main():
    """Entry point"""
    try:
        throttlebuster.add_command(download_command, "download")
        throttlebuster.add_command(estimate_command, "estimate")
        sys.exit(throttlebuster())

    except Exception as e:
        exception_msg = str({e.args[1] if e.args and len(e.args) > 1 else e})

        if DEBUG:
            logging.exception(e)
        else:
            if bool(exception_msg):
                logging.error(exception_msg)
            # sys.exit(show_any_help(e, exception_msg))

        sys.exit(1)



import os

import pytest

from tests import DOWNLOAD_DIR, FILE_URL, PART_DIR
from throttlebuster import DownloadedFile, DownloadMode, ThrottleBuster


def test_download_test():
    throttlebuster = ThrottleBuster(dir=DOWNLOAD_DIR, part_dir=PART_DIR)
    response = throttlebuster.run_sync(FILE_URL, test=True)
    assert response.is_success


@pytest.mark.parametrize(
    argnames=["tasks_amount"],
    argvalues=[
        (1,),
        (2,),
        (3,),
    ],
)
def test_real_download(tasks_amount: int):
    throttlebuster = ThrottleBuster(
        dir=DOWNLOAD_DIR,
        part_dir=PART_DIR,
        tasks=tasks_amount,
    )
    downloaded_file = throttlebuster.run_sync(FILE_URL, mode=DownloadMode.START, disable_progress_bar=True)
    assert downloaded_file.saved_to.exists()
    assert downloaded_file.tasks_used == tasks_amount
    assert downloaded_file.is_complete


def test_different_tasks_time():
    downloaded_file_items: list[DownloadedFile] = []

    for task in range(1, 5):
        # Ensure file size is big enough or throttling is small enough
        # for time difference to be noticed
        throttlebuster = ThrottleBuster(dir=DOWNLOAD_DIR, part_dir=PART_DIR, tasks=task)
        downloaded_file = throttlebuster.run_sync(
            FILE_URL, mode=DownloadMode.START, disable_progress_bar=True
        )
        assert downloaded_file.is_complete

        for file_part in downloaded_file.file_parts:
            assert file_part.is_complete

        if downloaded_file_items:
            previous_dowloaded_file = downloaded_file_items[-1]
            assert downloaded_file.duration < previous_dowloaded_file.duration

        downloaded_file_items.append(downloaded_file)

        os.remove(downloaded_file.saved_to)



"""
This library is designed to accelerate file downloads by overcoming
common throttling restrictions aiming to reduce download period
for large files

```python

from throttlebuster import ThrottleBuster

async def main():
    throttlebuster = ThrottleBuster()
    downloaded_file = await throttlebuster.run(
        "http://localhost:8888/test.1.opus",
    )
    print(
        downloaded_file
    )

if __name__ == "__main__":
    import asyncio

    asyncio.run(main())

```

## Progress hook

```python
from throttlebuster import DownloadTracker, ThrottleBuster


async def callback_function(data: DownloadTracker):
    percent = (data.downloaded_size / data.expected_size) * 100
    print(f"> Downloading {data.saved_to.name} {percent:.2f}%", end="\r")


async def main():
    throttlebuster = ThrottleBuster(threads=1)
    downloaded_file await throttlebuster.run(
        "http://localhost:8888/test.1.opus", progress_hook=callback_function
    )
    print(
        downloaded_file
    )


if __name__ == "__main__":
    import asyncio

    asyncio.run(main())

```

## Synchronous

```python
from throttlebuster import ThrottleBuster

throttlebuster = ThrottleBuster()

downloaed_file = throttlebuster.run_sync("http://localhost:8888/test.1.opus")

print(
    downloaded_file
)

```
"""

from importlib import metadata

try:
    __version__ = metadata.version("throttlebuster")
except metadata.PackageNotFoundError:
    __version__ = "0.0.0"

__author__ = "Smartwa"
__repo__ = "https://github.com/Simatwa/throttlebuster"


from throttlebuster.constants import DownloadMode
from throttlebuster.core import ThrottleBuster
from throttlebuster.helpers import logger
from throttlebuster.models import DownloadedFile, DownloadTracker

__all__ = ["ThrottleBuster", "DownloadTracker", "DownloadedFile", "DownloadMode", "logger"]



"""Supportive functions"""

import logging
import re
import typing as t
import warnings

import tqdm

from throttlebuster.constants import ILLEGAL_CHARACTERS_PATTERN

logger = logging.getLogger(__name__)

warnings.simplefilter("ignore", category=(tqdm.std.TqdmWarning,))
"""Raised due to frac*"""


class CustomTqdm(tqdm.tqdm):
    """Formats `n` and `total` bar values"""

    @staticmethod
    def prettify_number(number: float) -> float:
        return round(number, 2)

    @property
    def format_dict(self) -> dict:
        prev_dict = super().format_dict
        prev_dict["n"] = self.prettify_number(prev_dict.get("n"))
        prev_dict["total"] = self.prettify_number(prev_dict.get("total"))
        return prev_dict


class DownloadUtils:
    @classmethod
    def bytes_to_mb(self, bytes: int) -> int:
        return round(bytes / 1_000_000, 6)

    @classmethod
    def get_offset_load(cls, content_length: int, tasks: int) -> list[tuple[int, int]]:
        """Determines the bytes offset and the download size of each task

        Args:
            content_length (int): The size of file to be downloaded in bytes.
            tasks (int): Number of tasks for running the download.

        Returns:
            list[tuple[int, int]]: list of byte offset and download size for each task
        """
        assert tasks > 0, f"Threads value {tasks} should be at least 1"
        assert content_length > 0, f"Content-length value {content_length} should be more than 0"
        assert tasks < content_length, (
            f"Threads amount {tasks} should not be more than content_length {content_length}"
        )

        # Calculate base size and distribute remainder to the first few chunks
        base_size = content_length // tasks
        remainder = content_length % tasks
        load = [base_size + (1 if i < remainder else 0) for i in range(tasks)]

        assert sum(load) == content_length, "Chunk sizes don't add up to total length"
        assert len(load) == tasks, "Wrong number of chunks generated"

        # Generate (start_offset, chunk_size) pairs
        offset_load_container: list[tuple[int, int]] = []
        start = 0
        for size in load:
            offset_load_container.append((start, size))
            start += size

        return offset_load_container

    @classmethod
    def get_filename_from_header(cls, headers: dict) -> str | None:
        """Extracts filename from httpx response headers

        Args:
            headers (dict): Httpx response headers

        Returns:
            str | None: Extracted filename or None
        """
        disposition: str = headers.get("content-disposition")
        if disposition is not None:
            _, filename = disposition.split("filename=")
            return filename


def assert_instance(obj: object, class_or_tuple, name: str = "Parameter") -> t.NoReturn:
    """assert obj an instance of class_or_tuple"""

    assert isinstance(obj, class_or_tuple), (
        f"{name} value needs to be an instance of/any of {class_or_tuple} not {type(obj)}"
    )


def get_filesize_string(size_in_bytes: int) -> str:
    """Get something like `343 MB` or `1.25 GB` depending on size_in_bytes."""
    units = ["B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"]
    for unit in units:
        # 1024 or 1000 ?
        if size_in_bytes >= 1000.0:
            size_in_bytes /= 1000.0
        else:
            break
    return f"{size_in_bytes:.2f} {unit}"


def get_duration_string(time_in_seconds: int) -> str:
    """Get something like `2 Mins` or `3 Secs` depending on time_in_seconds."""
    units = ["Secs", "Mins", "Hrs"]
    for unit in units:
        if time_in_seconds >= 60.0:
            time_in_seconds /= 60.0
        else:
            break
    return f"{time_in_seconds:.2f} {unit}"


def sanitize_filename(filename: str) -> str:
    """Remove illegal characters from a filename"""
    return re.sub(ILLEGAL_CHARACTERS_PATTERN, "", filename.replace(":", "-"))



"""Constant variables"""

import os
import re
from enum import StrEnum
from pathlib import Path

CURRENT_WORKING_DIR = Path(os.getcwd())


DEFAULT_REQUEST_HEADERS = {
    "Accept": "*/",
    "User-Agent": "Mozilla/5.0 (X11; Linux x86_64; rv:137.0) Gecko/20100101 Firefox/137.0",
}

DEFAULT_REQUEST_COOKIES = {}

DOWNLOAD_PART_EXTENSION = ".part"

DEFAULT_CHUNK_SIZE = 256
"""In kilobytes"""

DEFAULT_TASKS = 2

DEFAULT_TASKS_LIMIT = 1000

DEFAULT_READ_TIMEOUT_ATTEMPTS = 10

ILLEGAL_CHARACTERS_PATTERN = re.compile(r"[^\w\-_\.\s()&|]")


class DownloadMode(StrEnum):
    START = "start"
    RESUME = "resume"
    AUTO = "auto"

    @classmethod
    def map(cls) -> dict:
        return {"start": cls.START, "resume": cls.RESUME, "auto": cls.AUTO}



"""Main module"""

import asyncio
import os
import shutil
import time
from pathlib import Path
from urllib.parse import urlparse

import aiofiles
import httpx
from httpx._types import HeaderTypes

from throttlebuster.constants import (
    CURRENT_WORKING_DIR,
    DEFAULT_CHUNK_SIZE,
    DEFAULT_READ_TIMEOUT_ATTEMPTS,
    DEFAULT_REQUEST_HEADERS,
    DEFAULT_TASKS,
    DEFAULT_TASKS_LIMIT,
    DOWNLOAD_PART_EXTENSION,
    DownloadMode,
)
from throttlebuster.exceptions import (
    FilenameNotFoundError,
    FilesizeNotFoundError,
    IncompatibleServerError,
)
from throttlebuster.helpers import (
    CustomTqdm,
    DownloadUtils,
    assert_instance,
    get_filesize_string,
    logger,
    sanitize_filename,
)
from throttlebuster.models import DownloadedFile, DownloadTracker


class ThrottleBuster(DownloadUtils):
    """Performs file download using multiple tasks in attempt
    to bypass the throttling limit. The download time `(t)` reduces
    to a new value `(nt)` by value `nt = t / th` where `(th)` is the
    tasks amount.

    This will only be useful when the throttling is done per `download
    stream` and NOT `per IP address` and server supports resuming download.

    #### NOTE
    Remember that increasing the number of concurrent tasks beyond what your
    system can handle efficiently may lead to performance degradation rather
    than improvement
    """

    tasks_limit: int = DEFAULT_TASKS_LIMIT
    """Number of tasks not to exceed"""

    def __init__(
        self,
        dir: Path | str = CURRENT_WORKING_DIR,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        tasks: int = DEFAULT_TASKS,
        part_dir: Path | str = CURRENT_WORKING_DIR,
        part_extension: str = DOWNLOAD_PART_EXTENSION,
        request_headers: HeaderTypes = DEFAULT_REQUEST_HEADERS,
        merge_buffer_size: int | None = None,
        **kwargs,
    ):
        """Constructor for `ThrottleBuster`

        Args:
            dir (Path | str, optional): Directory for saving downloaded files to. Defaults to CURRENT_WORKING_DIR.
            chunk_size (int, optional): Streaming download chunk size in kilobytes. Defaults to DEFAULT_CHUNK_SIZE.
            tasks (int, optional): Number of async tasks to carry out the download. Defaults to DEFAULT_TASKS.
            part_dir (Path | str, optional): Directory for temporarily saving downloaded file-parts to. Defaults to CURRENT_WORKING_DIR.
            part_extension (str, optional): Filename extension for download parts. Defaults to DOWNLOAD_PART_EXTENSION.
            request_headers (HeaderTypes, optional): Httpx request headers. Defaults to DEFAULT_REQUEST_HEADERS.
            merge_buffer_size (int|None, optional). Buffer size for merging the separated files in kilobytes. Defaults to chunk_size.

        kwargs : Keyword arguments for `httpx.AsyncClient`
        """  # noqa: E501
        # TODO: add temp-dir
        assert tasks > 0 and tasks <= self.tasks_limit, (
            f"Value for tasks should be atleast 1 and at most {self.tasks_limit}"
        )

        self.chunk_size: int = chunk_size * 1_024
        self.tasks: int = int(tasks)
        self.dir: Path = Path(dir)
        self.part_dir = Path(part_dir)
        self.part_extension: str = part_extension
        self.merge_buffer_size: int = (chunk_size if merge_buffer_size is None else merge_buffer_size) * 1_024
        self.client: httpx.AsyncClient = httpx.AsyncClient(**kwargs)
        """httpx AsyncClient"""
        self.client.headers.update(request_headers)

    def __repr__(self) -> str:
        return (
            rf"<{self.__module__}.{self.__class__.__name__} tasks={self.tasks} "
            rf'dir="{self.dir}", chunk_size_in_bytes={self.chunk_size}>'
        )

    def _generate_saved_to(self, filename: str, dir: Path, index: int | None = None) -> Path:
        filename, ext = os.path.splitext(filename)
        index = f".{index}" if index is not None else ""
        ext = ext if ext else ""

        return dir.joinpath(f"{filename}{index}{ext}")

    def _create_headers(self, bytes_offset: int) -> dict:
        new_headers = self.client.headers.copy()
        new_headers["Range"] = f"bytes={bytes_offset}-"
        return new_headers

    async def _call_progress_hook(self, progress_hook: callable, download_tracker: DownloadTracker) -> None:
        """Interacts with progress hook"""
        if progress_hook is None:
            return

        if asyncio.iscoroutinefunction(progress_hook):
            await progress_hook(download_tracker)

        else:
            progress_hook(download_tracker)
            # NOTE: Consider using status code to determine whether to proceed
            # with download process or not

    async def _merge_parts(
        self,
        file_parts: list[DownloadTracker],
        filename: Path,
        keep_parts: bool = False,
    ) -> Path:
        """Combines the separated download parts into one.

        Args:
            file_parts (list[DownloadTracker]): List of the separate files.
            filename (Path): Filename for saving the merged parts under.
            clear_parts (bool, optional): Whether to keep the download parts. Defaults to False.

        Returns:
            Path: Filepath to the merged parts.
        """

        save_to = self.dir.joinpath(filename)
        ordered_parts: list[DownloadTracker] = []

        for part in file_parts:
            assert part.saved_to.exists(), f"Part not found in downloaded path {part}"
            assert part.is_complete, f"Incomplete file part {part}"
            ordered_parts.insert(part.index, part)

        if len(file_parts) == 1:
            logger.info(f'Moving downloaded part to "{save_to}"')
            return Path(shutil.move(file_parts[0].saved_to, save_to))

        logger.info(f'Merging {len(file_parts)} file part{"s" if len(file_parts) > 1 else ""} to "{save_to}"')

        async with aiofiles.open(
            save_to,
            "wb",
        ) as fh:
            for part in ordered_parts:
                async with aiofiles.open(part.saved_to, "rb") as part_fh:
                    saved_size = 0
                    read_size = self.merge_buffer_size

                    while saved_size < part.expected_size:
                        chunk = await part_fh.read(min(read_size, part.expected_size - saved_size))
                        if not chunk:
                            break
                        await fh.write(chunk)
                        saved_size += len(chunk)

                if not keep_parts:
                    os.remove(part.saved_to)

        return save_to

    async def _downloader(
        self,
        download_tracker: DownloadTracker,
        progress_bar: CustomTqdm,
        progress_hook: callable,
    ) -> DownloadTracker:
        """Downloads each file part"""

        resume = False
        write_mode = "wb"

        match download_tracker.download_mode:
            case DownloadMode.AUTO:
                resume = download_tracker.saved_to.exists()

            case DownloadMode.RESUME:
                resume = download_tracker.saved_to.exists()

        if resume:
            write_mode = "ab"
            download_tracker.download_mode = DownloadMode.RESUME
            downloaded_size = os.path.getsize(download_tracker.saved_to)
            download_tracker.bytes_offset += downloaded_size
            download_tracker.update_downloaded_size(downloaded_size)
            progress_bar.n += self.bytes_to_mb(downloaded_size)
            progress_bar.last_print_t = time.time()
            progress_bar.last_print_n = progress_bar.n

        else:
            download_tracker.download_mode = DownloadMode.START

        logger.debug(
            f"Downloading file-part {download_tracker.index} "
            f"({get_filesize_string(download_tracker.downloaded_size)}/"
            f'{get_filesize_string(download_tracker.expected_size)}) to "{download_tracker.saved_to}" '
            f"resume - {resume}"
        )

        async with self.client.stream(
            "GET",
            url=download_tracker.url,
            headers=self._create_headers(download_tracker.bytes_offset),
        ) as stream:
            stream.raise_for_status()

            async with aiofiles.open(download_tracker.saved_to, write_mode) as fh:
                async for chunk in stream.aiter_bytes(self.chunk_size):
                    await fh.write(chunk)

                    download_tracker.update_downloaded_size(len(chunk))

                    progress_bar.update(self.bytes_to_mb(download_tracker.streaming_chunk_size))
                    await self._call_progress_hook(progress_hook, download_tracker)

                    if download_tracker.is_complete:
                        # Done downloading it's part
                        break

        return download_tracker

    async def run(
        self,
        url: str,
        filename: str = None,
        progress_hook: callable = None,
        mode: DownloadMode = DownloadMode.AUTO,
        disable_progress_bar: bool = None,
        file_size: int = None,
        keep_parts: bool = False,
        suppress_incompatible_error: bool = False,
        timeout_retry_attempts: int = DEFAULT_READ_TIMEOUT_ATTEMPTS,
        retry_attempts_count: int = 0,
        colour: str = "cyan",
        simple: bool = False,
        test: bool = False,
        leave: bool = True,
        ascii: bool = False,
        **kwargs,
    ) -> DownloadedFile | httpx.Response:
        """Initiate download process of a file.

        Args:
            url (str): Url of the file to be downloaded.
            filename (str, optional): Filename for the downloaded content. Defaults to None.
            progress_hook (callable, optional): Function to call with the download progress information. Defaults to None.
            mode (DownloadMode, optional): Whether to start or resume incomplete download. Defaults DownloadMode.AUTO.
            disable_progress_bar (bool, optional): Do not show progress_bar. Defaults to None (decide based on progress_hook).
            file_size (int, optional): Size of the file to be downloaded. Defaults to None.
            keep_parts (bool, optional): Whether to retain the separate download parts. Defaults to False.
            suppress_incompatible_error (bool, optional): Do no raise error when response headers lack Etag. Defaults to False.
            timeout_retry_attempts (int, optional): Number of times to retry download upon read request timing out. Defaults to DEFAULT_READ_TIMEOUT_ATTEMPTS.
            leave (bool, optional): Keep all leaves of the progressbar. Defaults to True.
            colour (str, optional): Progress bar display color. Defaults to "cyan".
            simple (bool, optional): Show percentage and bar only in progressbar. Deafults to False.
            test (bool, optional): Just test if download is possible but do not actually download. Defaults to False.
            ascii (bool, optional): Use unicode (smooth blocks) to fill the progress-bar meter. Defaults to False.

        kwargs: Other keyword arguments for `tqdm.tdqm`

        Returns:
            DownloadedFile | httpx.Response: Downloaded file details or httpx Response incase of (test=True).
        """  # noqa: E501

        assert_instance(mode, DownloadMode)

        if progress_hook is not None:
            assert callable(progress_hook), (
                f"Value for progress_hook must be a function not {type(progress_hook)}"
            )

        async_task_items = []
        download_tracker_items = []

        if disable_progress_bar is None:
            disable_progress_bar = progress_hook is not None

        logger.debug(f'Initializing download (tasks - {self.tasks}) for file in url - "{url}"')

        try:
            async with self.client.stream("GET", url=url) as stream:
                stream.raise_for_status()

                if (
                    stream.headers.get("Etag") is None
                    and suppress_incompatible_error is False
                    and self.tasks != 1
                ):
                    raise IncompatibleServerError(
                        "Server response header lacks Etag value which means "
                        "it doesn't support resuming downloads. "
                        "Set tasks to 1 or activate suppress_incompatible_error parameter "
                        "to silence this error."
                    )

                content_length = stream.headers.get("content-length", file_size)
                if type(content_length) is str:
                    content_length = int(content_length)

                filename = filename or self.get_filename_from_header(stream.headers)

                if filename is None:
                    # Try to get from path
                    _, filename = os.path.split(urlparse(url).path)
                    if not filename:
                        raise FilenameNotFoundError(
                            "Unable to get filename. Pass value using filename parameter "
                            "to suppress this error"
                        )

                filename = sanitize_filename(filename)

                final_saved_to = self._generate_saved_to(filename, self.dir)

                if content_length is None:
                    raise FilesizeNotFoundError(
                        "Unable to get the content-length of the file from server response. "
                        "Set the content-length using parameter file_size to suppres this error."
                    )

                elif not test and final_saved_to.exists() and mode is not DownloadMode.START:
                    if os.path.getsize(final_saved_to) == content_length:
                        logger.warning(f'Download already completed for the file in path "{final_saved_to}"')
                        return DownloadedFile(
                            url=url,
                            saved_to=final_saved_to,
                            size=os.path.getsize(final_saved_to),
                            duration=0,
                            file_parts=[],
                            merge_duration=0,
                            expected_size=content_length,
                        )

                size_with_unit = get_filesize_string(content_length)
                filename_disp = filename if len(filename) <= 8 + 3 else filename[:8] + "..."

                if test:
                    logger.info(f"Download test passed successfully ({size_with_unit}) - {final_saved_to}")
                    return stream

                logger.info(
                    f"{'Starting' if retry_attempts_count == 0 else 'Resuming'} "
                    f'download process ({self.tasks} tasks, {size_with_unit}) - "{filename}"'
                )
                p_bar = CustomTqdm(
                    total=self.bytes_to_mb(content_length),
                    desc=f"Downloading{f' [{filename_disp}]'}",
                    unit="Mb",
                    disable=disable_progress_bar,
                    colour=colour,
                    leave=leave,
                    ascii=ascii,
                    bar_format=(
                        "{l_bar}{bar} | %(size)s" % (dict(size=size_with_unit))
                        if simple
                        else "{l_bar}{bar}{r_bar}"
                    ),
                    **kwargs,
                )

                for index, offset_load in enumerate(self.get_offset_load(content_length, self.tasks)):
                    offset, load = offset_load
                    download_tracker = DownloadTracker(
                        url=url,
                        saved_to=self._generate_saved_to(
                            f"{filename}-{offset}{self.part_extension}",
                            self.part_dir,
                            index,
                        ),
                        index=index,
                        bytes_offset=offset,
                        expected_size=load,
                        download_mode=mode,
                    )

                    download_tracker_items.append(download_tracker)
                    async_task = asyncio.create_task(
                        self._downloader(
                            download_tracker,
                            progress_bar=p_bar,
                            progress_hook=progress_hook,
                        )
                    )
                    async_task_items.append(async_task)

                download_start_time = time.time()

                file_parts = await asyncio.gather(*async_task_items)

                download_duration = time.time() - download_start_time

                merge_start_time = time.time()
                p_bar.close()

                saved_to = await self._merge_parts(file_parts, filename=filename, keep_parts=keep_parts)

                downloaded_file = DownloadedFile(
                    url=url,
                    saved_to=saved_to,
                    expected_size=content_length,
                    size=os.path.getsize(saved_to),
                    file_parts=file_parts,
                    merge_duration=time.time() - merge_start_time,
                    duration=download_duration,
                )

                logger.info(
                    f"Done downloading {downloaded_file.size_string} "
                    f"in {downloaded_file.duration_string} "
                    f"{'merged' if self.tasks > 1 else 'moved'} in "
                    f"{downloaded_file.merge_duration_string}, "
                    f'saved to "{downloaded_file.saved_to}"'
                )

                return downloaded_file

        except httpx.ReadTimeout as e:
            retry_attempts_count += 1

            if retry_attempts_count <= timeout_retry_attempts:
                # Retry

                p_bar.close()

                logger.info(
                    f"Retrying download after read request timed out - "
                    f"attempt number ({retry_attempts_count}/{timeout_retry_attempts})"
                )

                return await self.run(
                    url=url,
                    filename=filename,
                    progress_hook=progress_hook,
                    mode=DownloadMode.AUTO,  # Changed
                    disable_progress_bar=disable_progress_bar,
                    file_size=file_size,
                    keep_parts=keep_parts,
                    suppress_incompatible_error=suppress_incompatible_error,
                    timeout_retry_attempts=timeout_retry_attempts,
                    retry_attempts_count=retry_attempts_count,
                    colour=colour,
                    simple=simple,
                    test=test,
                    leave=leave,
                    ascii=ascii,
                    **kwargs,
                )

            else:
                if timeout_retry_attempts:
                    logger.warning(
                        f"Giving up on download after exhausting all {timeout_retry_attempts} "
                        "read timeout retry attempts."
                    )

                else:
                    logger.info(
                        "Download read request has timed out. In order to automatically retry the "
                        "process, declare value for retry attempts using parameter "
                        "timeout_retry_attempts"
                    )

                raise e

    def run_sync(self, *args, **kwargs) -> DownloadedFile | httpx.Response:
        """Synchronously initiate download process of a file."""
        return asyncio.run(self.run(*args, **kwargs))



"""Error module"""


class ThrottleBusterException(Exception):
    """Base class for exceptions"""


class FilenameNotFoundError(ThrottleBusterException):
    """Raised when server response.headers does not contain `content-disposition`
    and user has not declared the filename value.
    """


class FilesizeNotFoundError(ThrottleBusterException):
    """Raised when server response.headers does not contain `content-length`
    and user has not declared the value for filename."""


class IncompatibleServerError(ThrottleBusterException):
    """Raised when server response lacks Etag in the headers"""



import subprocess

import pytest


def run_system_command(command: str) -> int:
    try:
        result = subprocess.run(
            "python -m throttlebuster " + command,
            shell=True,
            check=True,
            text=True,
            capture_output=True,
        )
        return result.returncode
    except subprocess.CalledProcessError as e:
        print(e, e.output, sep="\n")
        return e.returncode


def test_version():
    returncode = run_system_command("--version")
    assert returncode <= 0


@pytest.mark.parametrize(
    argnames=[
        "command",
    ],
    argvalues=[
        ["download --help"],
        ["estimate --help"],
    ],
)
def test_help(command):
    returncode = run_system_command(command)
    assert returncode <= 0


@pytest.mark.parametrize(
    argnames=[
        "command",
    ],
    argvalues=[
        ["download http://localhost:8888/miel-martin.webm --test"],
        ["download http://localhost:8888/miel-martin.webm --test --tasks 3"],
    ],
)
def test_download(command):
    returncode = run_system_command(command)
    assert returncode <= 0


@pytest.mark.parametrize(
    argnames=[
        "command",
    ],
    argvalues=[
        ["estimate --url http://localhost:8888/miel-martin.webm 500000"],
        ["estimate --size 1010029 300000"],
        ["estimate --url http://localhost:8888/miel-martin.webm 500000 --json"],
        ["estimate --size 1010029 300000 --json"],
    ],
)
def test_estimate(command):
    returncode = run_system_command(command)
    assert returncode <= 0



from throttlebuster import DownloadTracker, ThrottleBuster


async def callback_function(data: DownloadTracker):
    percent = (data.downloaded_size / data.expected_size) * 100
    print(f"> Downloading {data.saved_to.name} {percent:.2f}%", end="\r")


async def main(url: str):
    throttlebuster = ThrottleBuster(tasks=1)
    return await throttlebuster.run(
        url,
        # progress_hook=callback_function
    )


if __name__ == "__main__":
    import asyncio

    url = "http://localhost:8888/test.1.opus"

    out = asyncio.run(main(url))
    print(out)
